#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <getopt.h>
#include <pwd.h>
#include <grp.h>
#include <errno.h>
#include <syslog.h>
#include <iostream>
#include <memory>
#include <typeinfo>
using namespace std;
#include "kingate/fastcgi.h"
#include "kingate/cgi_gateway.h"
using namespace kingate;
#include "sitecing/sitecing_util.h"
#include "sitecing/util.h"
#include "sitecing/sitespace.h"
#include "sitecing/sitecing_interface_cgi.h"
#include "sitecing/cgi_component.h"
#include "sitecing/configuration.h"
#include "sitecing/magic.h"
#include "sitecing/sitecing_exception.h"
#include "sitecing/exception.h"
#include "sitecing/process_manager.h"
using namespace sitecing;

#include "config.h"
#define PHEADER PACKAGE " Version " VERSION
#define PCOPY   "Copyright (c) 2004 Klever Group"

class cdummyClass : public acomponent {
    public:
	void main(int _magic,va_list _args) {}
	void *__the_most_derived_this() { return NULL; }
} cdummyInstance;
class adummyClass : public cgi_component {
    public:
	void main(int _magic,va_list _args) {}
	void *__the_most_derived_this() { return NULL; }
} adummyInstance;

class sitecing_fastcgi_pm : public process_manager {
    public:
	configuration config;
	fcgi_socket *fss;
	semaphore sem;
	bool multi;
	uid_t uid;
	gid_t gid;
	pid_file pidfile;

	sitecing_fastcgi_pm(const string& config_file);
	~sitecing_fastcgi_pm();

	void process(int slot);

	void run();
	void give_up_privs();
};

sitecing_fastcgi_pm::sitecing_fastcgi_pm(const string& config_file)
    : config(config_file), multi(false) {
	if(( (config.flags&configuration::flag_user) || (config.flags&configuration::flag_group) || (config.flags&configuration::flag_chroot) ) && geteuid() )
	    throw konforka::exception(CODEPOINT,"can't use User, Group or Chroot when started without root privileges");
	if(config.flags&configuration::flag_user) {
	    struct passwd *ptmp = getpwnam(config.user.c_str());
	    if(ptmp) {
		uid = ptmp->pw_uid;
	    }else{
		errno=0;
		uid = strtol(config.user.c_str(),NULL,0);
		if(errno)
		    throw konforka::exception(CODEPOINT,"failed to resolve User value to uid");
	    }
	}
	if(config.flags&configuration::flag_group) {
	    struct group *gtmp = getgrnam(config.group.c_str());
	    if(gtmp) {
		gid = gtmp->gr_gid;
	    }else{
		errno=0;
		gid = strtol(config.group.c_str(),NULL,0);
		if(errno)
		    throw konforka::exception(CODEPOINT,"failed to resolve Group value to gid");
	    }
	}
	if(!(config.flags&configuration::flag_root_source))
	    throw konforka::exception(CODEPOINT,"unspecified root for sources");
	if(!(config.flags&configuration::flag_root_intermediate))
	    throw konforka::exception(CODEPOINT,"unspecified root for intermediate files");
	if(!(config.flags&configuration::flag_root_so))
	    throw konforka::exception(CODEPOINT,"unspecified root for shared objects");
	if(config.flags&configuration::flag_min_children)
	    min_children = config.min_children;
	if(config.flags&configuration::flag_max_children)
	    max_children = config.max_children;
	if(config.flags&configuration::flag_min_spare_children)
	    min_spare_children = config.min_spare_children;
	if(config.flags&configuration::flag_max_spare_children)
	    max_spare_children = config.max_spare_children;
	if(max_children<min_spare_children)
	    throw konforka::exception(CODEPOINT,"inconsistent numbers of MaxChildren and MinSpareChildren");
	if(min_children>max_spare_children && max_spare_children>=0)
	    throw konforka::exception(CODEPOINT,"inconsistent numbers of MinChildren and MaxSpareChildren");
	if(config.flags&configuration::flag_multi_process) {
	    multi = config.multi_process;
	}else{
	    if(config.flags&configuration::flag_listen_socket)
		multi = true;
	    else
		multi = false;
	}
	fss = (config.flags&configuration::flag_listen_socket)?new fcgi_socket(config.listen_socket.c_str(),5):new fcgi_socket(0);
	if(!fss)
	    throw konforka::exception(CODEPOINT,"failed to establish listening socket");
	if(config.flags&configuration::flag_daemonize && config.daemonize) {
	    pid_t pf = fork();
	    if(pf<0)
		throw konforka::exception(CODEPOINT,"failed to fork()");
	    if(pf) {
		die_humbly=true;
		_exit(0);
	    }
	}
	if(config.flags&configuration::flag_pid_file) {
	    pidfile.set(config.pid_file);
	}
	if(multi)
	    sem.init();
    }
sitecing_fastcgi_pm::~sitecing_fastcgi_pm() {
    if(fss)
	delete fss;
}

void sitecing_fastcgi_pm::process(int slot) {
    signal(SIGINT,SIG_DFL);
    signal(SIGABRT,SIG_DFL);
    signal(SIGTERM,SIG_DFL);
    give_up_privs();
    scoreboard_slot *sslot = sboard.get_slot(slot);
    try {
	sitespace ss(config);
	fcgi_socket& fs = *fss;
	sitecing_interface_cgi scif(&ss);
	string component_path;
	string action;
	config_options::action_handler_t *action_handler;
	int rpc = 0;
	if(config.flags&configuration::flag_requests_per_child)
	    rpc = config.requests_per_child;
	for(int req=0;(rpc<=0) || (req<rpc);rpc++) {
	    semaphore_lock sl;
	    if(multi) {
		sslot->state = scoreboard_slot::state_idle;
		sl.sem = &sem;
		sl.lock();
	    }
	    sslot->state = scoreboard_slot::state_accept;
	    fcgi_interface fi(fs);
	    sslot->state = scoreboard_slot::state_processing;
	    if(multi)
		sl.unlock();
	    cgi_gateway gw(fi);
	    scif.prepare(&gw);
	    try {
		component_path = normalize_path(gw.path_info(),strip_leading_slash|strip_trailing_slash);
		string full_component_path;
		while(true) {
		    full_component_path = config.root_source+'/'+component_path;
		    if(!access(full_component_path.c_str(),F_OK))
			break;
		    string::size_type sl = component_path.rfind('/');
		    if(sl==string::npos)
			throw konforka::exception(CODEPOINT,"can't find the target component");
		    component_path.erase(sl);
		}
		action = component_path;
		action_handler = config.lookup_action_handler(component_path);
		if(action_handler) {
		    action = action_handler->action;
		}
		string pwd = dir_name(full_component_path);
		if(chdir(pwd.c_str()))
		    throw konforka::exception(CODEPOINT,"failed to chdir() into document's directory");
		so_component soc = ss.fetch(action,&scif);
		if(action_handler) {
		    soc.ac->run(__magic_action,
			    config.root_source.c_str(), config.root_intermediate.c_str(), config.root_so.c_str(),
			    &(action_handler->args)
			    );
		}else{
		    soc.ac->main(0,NULL);
		}
	    }catch(http_status& hs) {
		scif.headers["Status"] = hs.status+" "+hs.message;
		string hshp = config.lookup_http_status_handler(component_path,hs.status);
		if(!hshp.empty()) {
		    so_component hsh = ss.fetch(hshp,&scif); // TODO: handle error trying to handle status
		    hsh.ac->run(__magic_http_status,config.root_source.c_str(),config.root_intermediate.c_str(),
			    config.root_so.c_str(),action.c_str(),&hs);
		}
	    }catch(compile_error& ce) {
		config_options *co_exception_handler = config.lookup_config(component_path,config_options::flag_exception_handler);
		if(co_exception_handler) {
		    so_component eh = ss.fetch(co_exception_handler->exception_handler,&scif); // TODO: handle error trying to handle error.
		    eh.ac->run(__magic_compile_error,ce.what(),config.root_source.c_str(),config.root_intermediate.c_str(),config.root_so.c_str(),ce.component_path.c_str());
		}else{
		    ce.see(CODEPOINT);
		    throw;
		}
	    }catch(preprocessor_error& pe) {
		config_options *co_exception_handler = config.lookup_config(component_path,config_options::flag_exception_handler);
		if(co_exception_handler) {
		    so_component eh = ss.fetch(co_exception_handler->exception_handler,&scif); // TODO: handle error trying to handle error.
		    eh.ac->run(__magic_preprocess_error,pe.what(),config.root_source.c_str(),config.root_intermediate.c_str(),config.root_so.c_str(),pe.component_name.c_str(),pe.line_number);
		}else{
		    pe.see(CODEPOINT);
		    throw;
		}
	    }catch(exception& e) {
		config_options *co_exception_handler = config.lookup_config(component_path,config_options::flag_exception_handler);
		if(co_exception_handler) {
		    so_component eh = ss.fetch(co_exception_handler->exception_handler,&scif); // TODO: handle error trying to handle error.
		    eh.ac->run(__magic_generic_exception,e.what(),config.root_source.c_str(),config.root_intermediate.c_str(),config.root_so.c_str(),component_path.c_str(),&e);
		}
	    }
	    scif.flush();
	}
    }catch(exception& e) {
	cerr << "->Oops: " << e.what() << endl;
    }
}

void sitecing_fastcgi_pm::run() {
    if(multi)
	manage();
    else
	process(0);
}

void sitecing_fastcgi_pm::give_up_privs() {
    if(config.flags&configuration::flag_chroot) {
	if(chroot(config.chroot.c_str()))
	    throw konforka::exception(CODEPOINT,"failed to chroot()");
    }
    if(config.flags&configuration::flag_group) {
	if((getgid()!=gid) && setgid(gid))
	    throw konforka::exception(CODEPOINT,"failed to setgid()");
    }
    if(config.flags&configuration::flag_user) {
	if((getuid()!=uid) && setuid(uid))
	    throw konforka::exception(CODEPOINT,"failed to setuid()");
    }
}

static sitecing_fastcgi_pm* _process_manager = NULL;

static void lethal_signal_handler(int signum) {
    _process_manager->finishing=true;
}

int main(int argc,char **argv) {
    const char* id = *argv;
    const char* t;
    while(t = index(id,'/')) {
	id=t; id++;
    }
    openlog(id,LOG_PERROR|LOG_PID,LOG_USER);
    try {
	string config_file = "sitecing.conf";
	while(true) {
	    static struct option opts[] = {
		{ "help", no_argument, 0, 'h' },
		{ "usage", no_argument, 0, 'h' },
		{ "version", no_argument, 0, 'V' },
		{ "license", no_argument, 0, 'L' },
		{ "config", required_argument, 0, 'f' },
		{ NULL, 0, 0, 0 }
	    };
	    int c = getopt_long(argc,argv,"f:hVL",opts,NULL);
	    if(c==-1)
		break;
	    switch(c) {
		case 'h':
		    cerr << PHEADER << endl
			<< PCOPY << endl << endl
			<< " -h, --help" << endl
			<< " --usage          display this text" << endl
			<< " -V, --version    display version number" << endl
			<< " -L, --license    show license" << endl
			<< " -f filename, --config=filename" << endl
			<< "                  specify configuration file to use" << endl;
		    exit(0);
		    break;
		case 'V':
		    cerr << VERSION << endl;
		    exit(0);
		    break;
		case 'L':
		    extern const char *COPYING;
		    cerr << COPYING << endl;
		    exit(0);
		    break;
		case 'f':
		    config_file = optarg;
		    break;
		default:
		    cerr << "Huh??" << endl;
		    break;
	    }
	}
	sitecing_fastcgi_pm sfpm(config_file);
	_process_manager = &sfpm;
	signal(SIGINT,lethal_signal_handler);
	signal(SIGABRT,lethal_signal_handler);
	signal(SIGTERM,lethal_signal_handler);
	sfpm.run();
    }catch(exception& e) {
	/* cerr << "Oops: " << e.what() << endl; */
	syslog(LOG_ERR,"uncaught exception: %s",e.what());
    }
    closelog();
}
